//  ratio.hpp  ---------------------------------------------------------------//

//  Copyright 2010 Vicente J. Botet Escriba

//  Distributed under the Boost Software License, Version 1.0.
//  See http://www.boost.org/LICENSE_1_0.txt


#ifndef BOOST_RATIO_HPP
#define BOOST_RATIO_HPP

#include <boost/ratio/include.hpp>

#endif  // BOOST_RATIO_HPP
